﻿// Eval2D Tutorial

// Feedback:
// Ben Carruthers
// ben@pandagears.com
// http://pandagears.com

// This tutorial will walk through code snippets that generate
// procedural textures. You can always skip to the end and start
// writing code that will be evaluated on the fly. All code here
// is saved when you exit the program.

// You can double-click the image view to maximize the Eval2D
// window, but keep in mind that evaluation may take longer if a
// larger image needs to be rendered.

// Here are some basic functions. To try them out, highlight each 
// line individually and press either F5 or Ctrl+Enter to evaluate
// the selected code. Keep in mind that the last statement with a
// return value is presented.
Func2.Distance.Multiply(1);
Func2.Noise.Multiply(1);
Func2.Voronoi();

// For the rest of the tutorial, you can follow the pattern 
// of highlighting a section of code and evaluating it alone to
// see the result. Try these lines to see how text output works.
Enumerable.Range(1, 5);
Matrix.Rotation(45).MultiplyPoint(10, 0);

// Here's where things get interesting. We can modify the input 
// or output of a function using (extension) methods available to
// the function. To increase the noise amplitude:
Func2.Noise.Multiply(3);

// Zoomed-out noise
// Scale() operates on the position values going into the noise
// function rather than on the output values.
Func2.Noise.Scale(5);

// Zoomed-in noise
// The trailing 'f' in 0.5f indicates this is a single-precision
// floating point value. It's inconvenient, but needed because
// the math library expects single-precision values.
Func2.Noise.Scale(0.5f);

// Noise mapped to [0, 1]
// Up to this point, sharp discontinuities have appeared in the
// noise function due to the way negative values are displayed.
// Generally, functions in the math library will output or expect
// values in the range [-1, 1], but when we want to visualize the
// final output, [0, 1] is needed.
Func2.Noise.Scale(5).To01();

// Noise mapped to [0, 1] using select
// Alternately, we can use Select(), passing in a lambda function
// that will define how an individual value will be mapped to a
// new output value (which can be anything, not necessarily a
// scalar). This approach is highly flexible, although common
// mappings such as To01() or Multiply() are defined already for
// convenience.
Func2.Noise.Scale(5).Select(s => (s + 1) / 2);

// Other simple functions
Func2.Noise.Scale(3).To01().Threshold(0.5f);
Func2.Voronoi().Scale(3).Threshold(0.25f).Invert01();
Func2.MinkowskyDistance(0.7f).Multiply(0.2f);

// Combining functions
// Sometimes we want to combine multiple functions using some
// operation on pairs of values. Here, we multiply noise by
// distance from center, then clamp the result to the range
// [0, 1]. Make sure to select all the lines together to 
// evaluate.
Func2.Combine(
	Func2.Noise.To01(),
	Func2.Distance.Invert01(),
	(a, b) => a * b)
	.Clamp01();
	
// Shorthand add
Func2.Add(
	Func2.DistanceSquared,
	Func2.Noise.To01())
	.Multiply(0.1f);
	
// Shorthand multiply
Func2.Multiply(
	Func2.DistanceSquared,
	Func2.Noise.To01())
	.Clamp01();

// Transformations
// We'll revisit transformations on input positions here.
// Scale() causes scaling on input position rather than on its
// output value. To multiply/scale output, use Multiply() instead.
// Similarly, Translate() is applied to input position and Add()
// to output values. Note that rotation angles are in degrees.
Func2.Noise.Rotate(30);
Func2.Noise.Rotate(60);
Func2.Noise.Translate(0.1f, 0);
Func2.Noise.Translate(0.2f, 0);

// Turbulent distance
// Here, we alter input positions by adding 2D noise values.
Func2.Distance  		// start with distance from center	
	.Invert01() 		// invert so center value is 1.0
    .Translate(			// apply turbulence   	
    	Func2.GradientNoise  	// start with 2D noise function
    		.Multiply(0.02f)	// reduce amplitude of noise
    		.Scale(4))			// finer granularity of noise
    .Scale(2);			// zoom out

// Radial distance arcs
// We create several distance functions and add them together.
// The Minkowsky distance calculation involves an exponent
// parameter which generalizes Euclidean distance (2.0),
// Manhattan distance (1.0), and Chebyshev distance (infinity).
// Also note the use of aggregate to repeatedly apply a pairwise
// operation, in this case addition.
Enumerable.Range(0, 3)
    .Select(i =>
        Func2.MinkowskyDistance(0.6f)
            .Invert01()
            .Rotate(i * 30))
    .Aggregate(Func2.Zero, Func2.Add);
    
// Seamless repeating/tiling noise
Func2.RepeatingNoise(4).Scale(3).To01();

// Fractional Brownian motion
// Here, Fbm() is used to successively sample four octaves of 
// noise, adding the values together. Note that the repeated 
// addition causes the output values to go beyond the range
// [-1, 1], so we clamp to this range before mapping linearly to
// [0, 1].
Func2.Noise.Fbm(4).Scale(4).Clamp().To01();

// Spiral with FBM
Func2.Spiral(0.2f).Fbm(6).Multiply(1.5f);

// Linear color gradient
// We take the X-coordinate of the sampled position and map it to
// a color using an RGBA gradient defined by a list of gradient
// stops. Between stops, color is linearly interpolated. The
// maximum offset is defined by the first parameter (10).
Func2.Position
	.Select(p => p.X)
	.To01()
	.Select(
		Func1.RgbaMap(10,	// define mapping from
			new uint[,]		// scalar output values
	        { 				// to RGBA values
	            {  0, 0x00000000 },
	            {  3, 0x221100ff },
	            {  7, 0xffaa77ff },
	            { 10, 0xffff99ff },
	        }));

// Voronoi/Worley noise with color
Func2.Voronoi()		// start with Voronoi defaults
	.Scale(5)		// zoom out	
	.Select(
		Func1.RgbaMap(10,	// define mapping from
			new uint[,]		// scalar output values
	        { 				// to RGBA values
	            {  0, 0x00000000 },
	            {  3, 0x221100ff },
	            {  7, 0xffaa77ff },
	            { 10, 0xffff99ff },
	        }));

// From here, we'll just define a variable for the
// color map to avoid redefining it for each function.
var map = Func1.RgbaMap(10,
	new uint[,]
	{ 
	    {  0, 0x000000ff },
	    {  6, 0xff0000ff },
	    {  7, 0xff4400ff },
	    {  8, 0xffaa00ff },
	    { 10, 0xffff00ff },
	});

// Some Musgrave noise variants
Func2.Noise.Multifractal(2, 2, 1, 0);
Func2.Noise.HybridMultifractal(3, 2, 1, 0);

// Voronoi Manhattan F2
// F2 indicates we're using the distance to the second
// nearest point rather than the nearest, which would
// be F1. Manhattan refers to Minkowsky distance with
// exponent 1.0 rather than Euclidean, which has exponent
// 2.0.
Func2.Voronoi((f1, f2) => f2, 1, 0.5f)
	.Scale(4)
	.Multiply(0.5f);

// Starfield
var n = Func2.Noise.To01().Multiply(20).Scale(5);
Func2.GradientNoise.Voronoi(0.6f, 1f)
 	.Scale(3)
 	.Clamp()
 	.Invert01()
 	.Select((s, p) => Scalar.Pow(s, n(p) + 15))
 	.Multiply(1f);
 	
// Magma
Func2.Noise
    .HeteroTerrain(4, 2f, 0.2f, 0.4f)
    .Scale(3)
    .Abs()
    .Multiply(0.2f)
    .Clamp01()
    .Select(map);

// Turbulent noise
Func2.Noise
	.Fbm(4)
    .Translate(
    	Func2.Noise
			.Fbm(1.5f)
			.Scale(3)
			.Multiply(0.3f)
			.Select(s => Vector.All(s)))
    .Scale(1)
    .Clamp()
    .To01();
 	
// Cell appearance noise
// First, we use a noise function that returns the point (vector)
// of the nearest Voronoi cell site. Next, we want to map these
// points to a scalar value so we can distinguish cell regions
// which share a common site.
Func2.VoronoiCell()
	.Scale(20) // zoom out
	.Select(Func2.Noise.Scale(0.1f)) // need to unscale zoom here
	.To01()
	.Select(map);
	
// Repeating pattern with warping
// Using a repeating function as a basis for Voronoi ensures that
// the Voronoi function also repeats.
Func2.RepeatingGradientNoise(5)
	.Voronoi((f1, f2) => f1 - f2 - 0.1f, 1.5f, 0.7f)	
	.RidgedMultifractal(4, 3, 0.8f, 0f)
	.Clamp01()
	.Scale(5)
	.Translate(Func2.GradientNoise.Scale(2).Multiply(0.01f));

// Some Musgrave noise variations below.

Func2.Noise.VariableLacunarity(0.4f).To01();

Func2.Noise
	.VariableLacunarity(Func2.GradientNoise.Scale(3), 0.5f)
	.To01();

// Parameters for the following are:
//   octaves
//   lacunarity
//   H (fractal increment)
//   offset
//   gain (ridged only)

Func2.Noise
	.HybridMultifractal(4, 3, 0.7f, 1)
	.Multiply(0.3f)
	.Scale(2);
	
Func2.Noise
	.RidgedMultifractal(3, 2, 0.5f, 1)
	.Translate(10, 10)
	.Multiply(0.4f)	
	.Select(Func1.RgbaMap(10,
		new uint[,]
		{ 
		    {  0, 0x000000ff },
		    {  3, 0x550000ff },
		    {  7, 0xff0000ff },
		    {  8, 0xff5533ff },
		    { 10, 0xffff55ff },
		}));
		
Func2.Noise.Multifractal(4, 3, 0.7f, 3f).Multiply(0.5f);

Func2.Noise
	.HeteroTerrain(6, 3, 0.5f, 0.5f)
	.Abs()
	.Multiply(0.2f)
	.Clamp01()
	.Pow(0.5f)
	.Scale(2)
	.Select(Func1.RgbaMap(10,
		new uint[,]
		{ 
		    {  0, 0x000000ff },
		    {  3, 0x112222ff },
		    {  7, 0x228844ff },
		    {  8, 0x44aa88ff },
		    { 10, 0x66ffccff },
		}));

// Write your own functions down here. If you come up with
// anything interesting, please share!